@@ -15,24 +15,27 @@ import atexit
 import datetime
 import fnmatch
 import gzip
-import hashlib
 import io
 import itertools
 import json
 import os
 import platform
+import queue
 import shutil
 import subprocess
 import sys
 import threading
 import typing as t
 import uuid
+from contextlib import contextmanager
 from copy import deepcopy
 from enum import Enum
 from hashlib import md5, sha1
 from pathlib import Path
 
+import dynaconf.utils
 import fsspec
+from doit.cmd_base import CmdAction
 from doit.cmd_base import DoitCmdBase as AltoCmdBase
 from doit.cmd_base import TaskLoader2 as DoitEngine
 from doit.loader import generate_tasks
@@ -62,7 +65,7 @@ from alto.utils import (
 )
 
 if t.TYPE_CHECKING:
-    from dynaconf import Validator
+    from dynaconf import Dynaconf, Validator
 
 __all__ = [
     "AltoConfiguration",
@@ -77,6 +80,28 @@ RESERVOIR_BUFFER_SIZE = 10_000
 """The default number of records to buffer before flushing to reservoir filesystem."""
 
 
+def find_hyphen_key(key: str, data: t.Dict[str, t.Any]) -> t.Optional[str]:
+    """Find a key in a dict that matches the given key, allowing for hyphens."""
+    if key in data:
+        return key
+    for k in data.keys():
+        if k.replace("-", "_") == key:
+            return k
+    return None
+
+
+# Patch dynaconf to support hyphenated keys
+# this allows us to override config values with env vars
+# even when the config key contains a hyphen.
+# ie. `ALTO_TAPS__MY_TAP__SOMETHING` will override `taps.my-tap.something`
+# If the hypenated key is not found, the original dynaconf lookup is used.
+# This means non-hyphenated keys will still work as expected if they exist.
+__case_lookup = dynaconf.utils.find_the_correct_casing
+dynaconf.utils.find_the_correct_casing = lambda key, data: find_hyphen_key(
+    key, data
+) or __case_lookup(key, data)
+
+
 class AltoCmd(str, Enum):
     """The alto task type is used to determine the type of task to execute."""
 
@@ -256,7 +281,7 @@ class AltoFileSystem:
                 self._fs = DirFileSystem(
                     f"{self.config['BUCKET']}/{path}/{self.config['PROJECT_NAME']}",
                     fs=fsspec.filesystem(
-                        fsystem.lower(), **self.config.get(f"{fsystem}_KWARGS", DynaBox())
+                        fsystem.lower(), **self.config.get(f"{fsystem}_SETTINGS", DynaBox())
                     ),
                 )
             else:
@@ -333,7 +358,9 @@ class AltoFileSystem:
             remote: Whether or not the path should be in the remote storage directory.
         """
         getter = self._remote_path if remote else self._temp_path
-        return getter(fname=f"{tap}-to-{target}.json", key=STATE_DIR.format(env=self.config["ENV"]))
+        return getter(
+            fname=f"{tap}-to-{target}.json", key=STATE_DIR.format(env=self.config.current_env)
+        )
 
     def base_catalog_path(self, name: str, remote: bool = False) -> str:
         """Return the path to the base catalog for a plugin.
@@ -365,7 +392,7 @@ class AltoFileSystem:
             remote: Whether or not the path should be in the remote storage directory.
         """
         getter = self._remote_path if remote else self._root_path
-        return getter(fname=fname, key=LOG_DIR.format(env=self.config["ENV"]))
+        return getter(fname=fname, key=LOG_DIR.format(env=self.config.current_env))
 
 
 class AltoPlugin:
@@ -377,6 +404,9 @@ class AltoPlugin:
         self.type = typ
         self.alto = config
         self._spec = self.alto.spec_for(name)
+        # Permit dynamic select override local to the cls instance
+        self._select = None
+        self._retain_hash_rules = True
 
     @property
     def spec(self) -> DynaBox:
@@ -390,7 +420,10 @@ class AltoPlugin:
             # Utility plugins can optionally specify a pip_url
             return self.spec.get("pip_url")
         # All other plugins must specify a pip_url
-        return self.spec["pip_url"]
+        try:
+            return self.spec["pip_url"]
+        except KeyError:
+            raise KeyError(f"Plugin {self.name} is missing a pip_url.")
 
     @property
     def parent(self) -> t.Optional["AltoPlugin"]:
@@ -410,7 +443,10 @@ class AltoPlugin:
     @property
     def root_namespace(self) -> str:
         """Return the root namespace for the plugin."""
-        return self.alto.inner["LOAD_PATH"]
+        try:
+            return self.alto.inner["LOAD_PATH"]
+        except KeyError:
+            raise KeyError("Alto is missing a top-level LOAD_PATH.")
 
     @property
     def namespace(self) -> str:
@@ -455,7 +491,19 @@ class AltoPlugin:
     @property
     def select(self) -> t.List[str]:
         """Return the select for the plugin."""
-        return self.spec.get("select", ["*.*"])
+        if self._select is None:
+            self._select = self.spec.get("select", ["*.*"])
+        return self._select
+
+    @select.setter
+    def select(self, value: t.List[str]) -> None:
+        """Set the select for the plugin."""
+        # If we're setting the select, we want to retain any hash rules
+        # as PII hashing is a global concern and it should be on the user
+        # to _explicitly_ disable it.
+        self._select = (value or ["*.*"]) + [
+            rule for rule in self.select if rule.startswith("~") and self._retain_hash_rules
+        ]
 
     @property
     def metadata(self) -> t.Dict[str, t.Any]:
@@ -471,7 +519,11 @@ class AltoPlugin:
     def environment(self) -> str:
         """Return env vars necessary to run our PEX executable."""
         typ = "MODULE" if "entrypoint" in self.spec else "SCRIPT"
-        return {f"PEX_{typ}": self.entrypoint, "ALTO_PLUGIN": self.name}
+        return {
+            f"PEX_{typ}": self.entrypoint,
+            "ALTO_PLUGIN": self.name,
+            **self.spec.get("environment", {}),
+        }
 
     def config_relative_to(self, other: "AltoPlugin") -> DynaBox:
         """Return the config for the plugin."""
@@ -594,9 +646,9 @@ class AltoStreamMap:
         """
         if value.get("type") == "object":
             for k, v in value.get("properties", {}).items():
-                self.recursive_schema_apply(v, f"{crumb}.{k}")
+                self.recursive_schema_apply(v, f"{crumb}.{k}", transformer)
         elif value.get("type") == "array":
-            self.recursive_schema_apply(value["items"], crumb)
+            self.recursive_schema_apply(value["items"], crumb, transformer)
         elif self.crumb_selected(crumb):
             value = transformer(value)
         return value
@@ -611,9 +663,12 @@ class AltoStreamMap:
         implementing their own stream maps.
         """
         if isinstance(value, dict):
-            return {k: self.recursive_record_apply(v, f"{crumb}.{k}") for k, v in value.items()}
+            return {
+                k: self.recursive_record_apply(v, f"{crumb}.{k}", transformer)
+                for k, v in value.items()
+            }
         if isinstance(value, list):
-            return [self.recursive_record_apply(v, crumb) for v in value]
+            return [self.recursive_record_apply(v, crumb, transformer) for v in value]
         elif self.crumb_selected(crumb):
             return transformer(value)
         return value
@@ -679,7 +734,7 @@ class HashStreamMap(AltoStreamMap):
 
     def _pii_hash(self, value: t.Any) -> t.Any:
         """Hash a value."""
-        return hashlib.md5(str(value).encode("utf-8"), usedforsecurity=False).hexdigest()
+        return md5(str(value).encode("utf-8")).hexdigest()
 
     def transform_record(self, record: dict) -> dict:
         """Transform the record."""
@@ -753,6 +808,17 @@ class AltoExtension:
         """Return the tasks for the extension."""
         yield from ()
 
+    def _tasks(self) -> t.Generator[AltoTaskData, None, None]:
+        """Return the tasks for the extension.
+
+        Not used but exists as reference to a good pattern.
+        """
+        for task in self.tasks():
+            for i, action in enumerate(task["actions"]):
+                if isinstance(action, str):
+                    task["actions"][i] = CmdAction(action, env=os.environ)
+            yield task
+
 
 class AltoTaskEngine(DoitEngine):
     """The alto task engine builds on top of doit to provide a simple interface for Singer tasks.
@@ -767,21 +833,54 @@ class AltoTaskEngine(DoitEngine):
     def __init__(
         self,
         root_dir: Path = Path.cwd(),
+        config: t.Optional[t.Union["Dynaconf", t.Dict[str, t.Any]]] = None,
     ) -> None:
-        """Initialize the alto task engine."""
+        """Initialize the alto task engine.
+
+        Args:
+            root_dir: The root directory of the project. This is where the configuration is
+                loaded from if no configuration is provided. Defaults to the current working
+                directory.
+            config: The configuration to use. If this is a dictionary, it will be written to a
+                temporary file and loaded. If this is a Dynaconf object, it will be used directly.
+                If this is None, the configuration will be loaded from the root directory.
+        """
         super().__init__()
         from dynaconf import Dynaconf
 
-        self.alto = Dynaconf(
-            settings_files=[root_dir / f"alto.{fmt}" for fmt in SUPPORTED_CONFIG_FORMATS],
-            secrets=[root_dir / f"alto.secrets.{fmt}" for fmt in SUPPORTED_CONFIG_FORMATS],
-            root_path=root_dir,
-            envvar_prefix="ALTO",
-            env_switcher="ALTO_ENV",
-            load_dotenv=True,
-            environments=True,
-            merge_enabled=True,
-        )
+        kwargs = {
+            "root_path": root_dir,
+            "envvar_prefix": "ALTO",
+            "env_switcher": "ALTO_ENV",
+            "load_dotenv": True,
+            "environments": True,
+            "merge_enabled": True,
+        }
+        if config is None:
+            # Default to loading the configuration from the root directory
+            self.alto = Dynaconf(
+                settings_files=[root_dir / f"alto.{fmt}" for fmt in SUPPORTED_CONFIG_FORMATS],
+                secrets=[
+                    root_dir / f"alto.{segment}.{fmt}"
+                    for fmt in SUPPORTED_CONFIG_FORMATS
+                    for segment in ("secrets", "local")
+                ],
+                **kwargs,
+            )
+        elif isinstance(config, Dynaconf):
+            # Use the provided configuration
+            self.alto = config
+        elif isinstance(config, dict):
+            # Write the configuration to a temporary file and load it
+            from tempfile import NamedTemporaryFile
+
+            with NamedTemporaryFile(mode="w", suffix=".json") as f:
+                f.write(json.dumps(config))
+                self.alto = Dynaconf(
+                    settings_files=[f.name],
+                    root_path=root_dir,
+                    **kwargs,
+                )
         # Instantiate the filesystem and configuration
         self.configuration = AltoTaskEngine.Configuration(inner=self.alto)
         self.filesystem = AltoTaskEngine.FileSystem(root_dir, config=self.configuration)
@@ -792,9 +891,16 @@ class AltoTaskEngine(DoitEngine):
         """Return the filesystem."""
         return self.filesystem.fs
 
-    def setup(self, opt_values: t.Dict[str, t.Any]) -> None:
+    def setup(self, opt_values: t.Optional[t.Dict[str, t.Any]] = None) -> None:
         """Load extensions and filesystem interface. This is called by doit."""
+        if opt_values is None:
+            opt_values = {}
+        # Set the environment variables
+        for k, v in self.alto.get("ENVIRONMENT", {}).items():
+            os.environ[k] = str(v)
+        # Load the extensions
         self._load_extensions()
+        # Validate the configuration
         self.alto.validators.validate_all()
 
     def _load_extensions(self) -> None:
@@ -1018,7 +1124,7 @@ class AltoTaskEngine(DoitEngine):
                     (get_remote_state, (tap_reservoir, target.name, self.filesystem, True)),
                     (
                         reservoir_to_target,
-                        (tap, target, pipeline_id, self.filesystem, self.alto["ENV"]),
+                        (tap, target, pipeline_id, self.filesystem, self.alto.current_env),
                     ),
                 )
                 .set_task_dep(f"{AltoCmd.BUILD}:{target}")
@@ -1056,7 +1162,7 @@ class AltoTaskEngine(DoitEngine):
                             tap,
                             pipeline_id,
                             self.filesystem,
-                            self.alto["ENV"],
+                            self.alto.current_env,
                             self.alto.get("RESERVOIR_BUFFER_SIZE", RESERVOIR_BUFFER_SIZE),
                         ),
                     ),
@@ -1102,7 +1208,7 @@ class AltoTaskEngine(DoitEngine):
                 generate_tasks(AltoCmd.ABOUT, self.task_about(), self.task_about.__doc__),
                 *(
                     generate_tasks(ext.name, ext.tasks(), f"[extension] {ext.__doc__}")
-                    for n, ext in enumerate(self.extensions)
+                    for ext in self.extensions
                 ),
             )
         )
@@ -1120,6 +1226,10 @@ def map_worker(
     for line in instream:
         if not line.strip():
             continue
+        if not mappers:
+            # no mappers, just pass through
+            outstream.write(line)
+            continue
         try:
             message = json.loads(line)
         except json.JSONDecodeError:
@@ -1133,7 +1243,8 @@ def map_worker(
         else:
             outstream.write(line)
             continue
-        outstream.write(json.dumps(message).encode("utf-8") + b"\n")
+        if message:
+            outstream.write(json.dumps(message).encode("utf-8") + b"\n")
 
 
 # ==================== #
@@ -1956,7 +2067,7 @@ def render_config(
     settings: t.Dict[str, t.Any],
     filesystem: AltoFileSystem,
     accent: t.Optional[AltoPlugin] = None,
-) -> None:
+) -> dict:
     """Render a config file for a plugin."""
     # Acquire the lock, this is necessary because the config rendering process
     # is not thread-safe.
@@ -1976,12 +2087,15 @@ def render_config(
         # Render the config
         config_path = Path(filesystem.config_path(plugin.name, accent.name if accent else None))
         config_path.parent.mkdir(parents=True, exist_ok=True)
+        runtime_config = config.to_dict()
         with open(config_path, "w") as f:
-            json.dump(config.to_dict(), f, indent=2)
+            json.dump(runtime_config, f, indent=2)
 
         # Reset the namespace
         settings["LOAD_PATH"] = original_namespace
 
+    return runtime_config
+
 
 # ============== #
 # PEX Management #
@@ -2047,3 +2161,199 @@ def maybe_remove_pex(plugin: AltoPlugin, filesystem: AltoFileSystem) -> bool:
     except Exception:
         raise RuntimeError(f"Could not remove {remote} from remote cache. It may not exist.")
     return True
+
+
+# ================== #
+# Python API Helpers #
+# ================== #
+
+
+def setup_tap_target(
+    tap_name: str,
+    target_name: str,
+    filesystem: AltoFileSystem,
+    configuration: AltoConfiguration,
+    settings: t.Dict[str, t.Any],
+) -> t.Tuple[AltoPlugin, AltoPlugin, t.Dict[str, dict]]:
+    """Setup a tap and target for execution.
+
+    This function will build the pexes for the tap and target, generate the
+    base catalog for the tap, and render the config files for both the tap
+    and target.
+    """
+    tap, target = make_plugins(
+        tap_name, target_name, filesystem=filesystem, configuration=configuration
+    )
+    if not maybe_get_catalog(tap, filesystem):
+        generate_catalog(tap, filesystem)
+    render_modified_catalog(tap, filesystem)
+    _lock = threading.Lock()
+    output_configs = {}
+    output_configs[tap_name] = render_config(tap, _lock, settings, filesystem)
+    output_configs[target_name] = render_config(target, _lock, settings, filesystem, tap)
+    return tap, target, output_configs
+
+
+def make_plugins(
+    *plugin_names: str,
+    filesystem: AltoFileSystem,
+    configuration: AltoConfiguration,
+) -> t.Tuple[AltoPlugin, ...]:
+    """Create a tuple of plugins from a list of plugin names.
+
+    This function will build the pexes for the plugins if they do not exist.
+    """
+    plugins = []
+    for plugin_name in plugin_names:
+        plugin = configuration.get_plugin(plugin_name)
+        if not maybe_get_pex(plugin, filesystem):
+            build_pex(plugin, filesystem)
+        plugins.append(plugin)
+    return tuple(plugins)
+
+
+class _QueueFile(io.BytesIO):
+    """A file-like object that writes to a queue.
+
+    This is only intended to be used inside the `tap_runner` function. It
+    expects a thread to be writing to the queue and a thread to be reading
+    from the queue via the `__iter__` and `__next__` methods. Instances of
+    this class should be used in a for loop to iterate over the records.
+    Some iterations can be None, these should be ignored. A StopIteration
+    exception will be raised when the queue is empty and no new records
+    have been added for `max_wait` seconds.
+    """
+
+    def __init__(self, max_wait: int = 5, records_only: bool = False):
+        super().__init__()
+        self.queue = queue.Queue()
+        # The maximum time to wait for a new record,
+        # this is used to prevent the tap from hanging indefinitely
+        self.max_wait = max_wait
+        # Whether to only return records or all output, when records_only is
+        # True, the output will be a tuple of (stream, record)
+        self.records_only = records_only
+        # Store state
+        self._state = {}
+
+    def write(self, data) -> int:
+        self.queue.put(data)
+        return len(data)
+
+    def __iter__(self):
+        return self
+
+    def __next__(self) -> t.Union[dict, t.Tuple[str, dict], None]:
+        try:
+            data = self.queue.get(timeout=self.max_wait)
+        except queue.Empty:
+            raise StopIteration
+        else:
+            try:
+                rv = json.loads(data)
+            except json.JSONDecodeError:
+                pass
+            else:
+                if "type" not in rv:
+                    pass
+                elif rv["type"] == "STATE":
+                    merge(rv["value"], self._state)
+                elif self.records_only and not rv["type"] == "RECORD":
+                    pass
+                elif self.records_only:
+                    return (rv["stream"], rv["record"])
+                else:
+                    return rv
+            finally:
+                self.queue.task_done()
+
+    def close(self) -> None:
+        self.queue.join()
+
+
+@contextmanager
+def tap_runner(
+    tap: AltoPlugin,
+    filesystem: AltoFileSystem,
+    settings: t.Dict[str, t.Any],
+    state_key: str,
+    max_wait: int = 5,
+    records_only: bool = False,
+) -> t.Generator[_QueueFile, None, None]:
+    """Run a tap and yield a file-like object that reads from the tap's stdout."""
+    tap_bin, tap_config, tap_catalog = (
+        filesystem.executable_path(tap.pex_name),
+        filesystem.config_path(tap.name),
+        filesystem.catalog_path(tap.name),
+    )
+    state = filesystem.state_path(tap.name, state_key)
+    cmd = [tap_bin, "--config", tap_config]
+    if os.path.isfile(state) and tap.supports_state:
+        cmd += ["--state", state]
+    if tap.supports_catalog:
+        cmd += ["--catalog", tap_catalog]
+    elif tap.supports_properties:
+        cmd += ["--properties", tap_catalog]
+    lock = threading.Lock()
+    mappers = tap.get_stream_maps(filesystem)
+    _tmp_id = str(uuid.uuid4())
+    # Config
+    render_config(tap, lock, settings, filesystem)
+    # Catalog
+    if not maybe_get_catalog(tap, filesystem):
+        generate_catalog(tap, filesystem)
+    render_modified_catalog(tap, filesystem)
+    # State
+    get_remote_state(tap.name, state_key, filesystem, execute=tap.supports_state)
+    with subprocess.Popen(
+        cmd,
+        stdout=subprocess.PIPE,
+        stderr=subprocess.PIPE,
+        env={**os.environ, **tap.environment},
+        cwd=filesystem.root_dir,
+    ) as tap_proc:
+        log_thread = threading.Thread(
+            target=pipe_logger,
+            args=(
+                tap_proc.stderr,
+                filesystem.log_path(f"tap-{_tmp_id}.log"),
+                lock,
+            ),
+            daemon=True,
+        )
+        log_thread.start()
+        _stream = _QueueFile(max_wait, records_only)
+        # The map_worker is smart enough to just pass through the data
+        # if there are no mappers. This queue is used to prevent
+        # the tap from hanging indefinitely on read.
+        map_thread = threading.Thread(
+            target=map_worker,
+            args=(tap_proc.stdout, _stream, mappers),
+            daemon=True,
+        )
+        map_thread.start()
+        # Yield the stream iterator
+        yield _stream
+        # Cleanup
+        tap_proc.terminate()
+        tap_proc.wait()
+        log_thread.join(), map_thread.join()
+        # Commit State
+        if tap.supports_state:
+            with open(state, "w") as f:
+                json.dump(_stream._state, f)
+            update_remote_state_no_stdout(tap.name, state_key, filesystem)
+
+
+def get_engine(env: str, root_dir: t.Optional[Path] = None) -> AltoTaskEngine:
+    """Instantiate an AltoTaskEngine.
+
+    This is a convenience function for use in Python scripts. The
+    engine is configured entirely from the alto configuration file.
+    """
+    if root_dir is None:
+        root_dir = Path.cwd()
+    os.environ["ALTO_ENV"] = env
+    engine = AltoTaskEngine(root_dir)
+    engine.setup()
+    return engine
